/* Copyright (c) 1992, 1998, 2000, 2002, 2003, 2004, 2005, 2006 John E. Davis
 * This file is part of JED editor library source.
 *
 * You may distribute this file under the terms the GNU General Public
 * License.  See the file COPYING for more information.
 */

#include "config.h"
#include "jed-feat.h"
/*{{{ Include Files */

#include <stdio.h>
#include <string.h>

#include "buffer.h"
#include "ins.h"
#include "line.h"
#include "paste.h"
#include "screen.h"
#include "misc.h"
#include "cmds.h"

/* SLang user object be larger than 128 */
#define JED_MARK_TYPE 128

Buffer *Paste_Buffer;
static Buffer *Rectangle_Buffer;

////
int clipBoard;

int N_ROW, W_COL, S_ROW, E_COL, PULSE_R, PULSE_C;
// rectangle borders
// saved from corresponding buffer variables
// to be available globally to jed interpreter
// changed everytime check_rectangle is called
// ticking corner identified by PULSE_R, PULSE_C

Line *BegL, *EndL;
int BegRow, BegPoint, EndRow, EndPoint;
//  start/end coordinates of character block

static int RestoreHilight=0;
// 0 restore char stream, 2 restore whole lines
// 1 for rectangles

int DeletionToKillArray=1;
// used to copy deleted region to kill_array
// except when region is cut to clipboard

Buffer *thisbuf=NULL;
// used to append or overwrite in kill_array

// protos
int check_rectangle(void);
int toggle_block (int *blktype);
int restore_rectangle (void);
void append_region_to_kill_array (int *);

////

/* This is used by the narrow command so that multiple windows are
 * handled properly. The assumption is that we are dealing with a canonical
 * region */

static void touch_windows(void)
{
   Window_Type *w;
   Line *l;
   unsigned int n;

   w = JWindow;
   JWindow = JWindow->next;
   while (JWindow != w)
     {
        if (CBuf == JWindow->buffer)
          {
             /* The mark is set with line at top of buffer */
             l = CBuf->marks->line;
             n = CBuf->marks->n;
             while ((l != NULL) && (l != JWindow->mark.line)) l = l->next, n++;
             if (l == NULL)
               jed_init_mark (&JWindow->mark, 0);
             touch_window();
          }
        JWindow = JWindow->next;
     }
   jed_pop_mark(0);
}

#if 0
static int line_in_buffer(Line *line)
{
   Line *l;

   l = CBuf->beg;
   while (l != NULL) if (l == line) return(1); else l = l->next;
   return(0);
}

#endif

/*{{{ Spot/Mark Functions */

void jed_copy_mark (Mark *dest, Mark *src)
{
   *dest = *src;
   dest->next = NULL;
}

void jed_init_mark (Mark *m, unsigned int flags)
{
   m->line = CLine;
   m->point = Point;
   m->n = LineNum + CBuf->nup;
   m->flags = flags;
   /* m->next = NULL; */
}

void jed_init_mark_for_buffer (Mark *m, Buffer *b, unsigned int flags)
{
   m->line = b->line;
   m->point = b->point;
   m->n = b->linenum + b->nup;
   m->flags = flags;
   /* m->next = NULL; */
}

#if 0
int jed_init_mark_for_line (Mark *m, Line *l, unsigned int flags)
{
   Line *ll;
   unsigned int n;

   n = 1;
   ll = CBuf->beg;
   while ((ll != NULL) && (ll != l))
     {
        ll = ll->next;
        n++;
     }
   if (ll == NULL)
     return -1;

   m->line = l;
   m->point = 0;
   m->n = n + CBuf->nup;
   m->flags = flags;
   /* m->next = NULL; */
   return 0;
}
#endif

#if JED_HAS_SAVE_NARROW
static Mark *create_mark (unsigned int flags)
{
   Mark *m;

   if (NULL == (m = (Mark *) jed_malloc0 (sizeof(Mark))))
     exit_error("create-mark: malloc error", 0);

   jed_init_mark (m, flags);

   return m;
}
#endif

/* with prefix argument, pop marks */
int set_mark_cmd (void)
{
   Mark *m;
   if (Repeat_Factor != NULL)
     {
        while (CBuf->marks != NULL) jed_pop_mark(0);
        Repeat_Factor = NULL;
        return(1);
     }

   if (CBuf->marks == NULL)
     {
        jed_push_mark ();
     }

   m = CBuf->marks;

   jed_init_mark (m, m->flags);

   if ((m->flags & VISIBLE_MARK) == 0)
     {
        m->flags |= VISIBLE_MARK;
        CBuf->vis_marks++;
     }

   /* if (m != CBuf->marks) m->next = CBuf->marks;
    CBuf->marks = m; */
   if (Last_Key_Function == (FVOID_STAR) set_mark_cmd) message("Mark Set.");
   return(1);
}

static Mark *allocate_mark (Jed_Mark_Array_Type **ap, Mark **mp, unsigned int flags)
{
   Jed_Mark_Array_Type *a, *b;
   Mark *m;

   a = *ap;
   if ((a == NULL)
       || (a->num_marks == JED_MAX_MARK_ARRAY_SIZE))
     {
        if (NULL == (b = (Jed_Mark_Array_Type *) jed_malloc0 (sizeof(Jed_Mark_Array_Type))))
          {
             exit_error("allocate_mark: malloc error", 0);
          }
        b->next = a;
        a = b;
        *ap = a;
     }

   m = &a->marks[a->num_marks];
   jed_init_mark (m, flags);

   a->num_marks += 1;
   m->next = *mp;
   *mp = m;
   return m;
}

static void deallocate_mark (Jed_Mark_Array_Type **ap, Mark **mp)
{
   Jed_Mark_Array_Type *a;
   Mark *m;

   m = *mp;
   *mp = m->next;

   a = *ap;
   a->num_marks -= 1;

   if (a->num_marks == 0)
     {
        *ap = a->next;
        SLfree ((char *)a);
     }
}

int push_spot()
{
   (void) allocate_mark (&CBuf->spot_array, &CBuf->spots, 0);
   return 1;
}

int jed_push_mark()
{
   (void) allocate_mark (&CBuf->mark_array, &CBuf->marks, 0);
   return 1;
}

int jed_goto_mark (Mark * m)
{
    Line *l;
    int ret = -1;

    l = m->line;
    LineNum = m->n;

    if (LineNum <= CBuf->nup) bob();
    else if (LineNum > CBuf->nup + Max_LineNum) eob();
    else
    {
        CLine = l;
        Point = m->point;
        LineNum -= CBuf->nup;
        ret = 0;
    }
    return ret;
}

int jed_pop_mark (int go)
{
   Mark *m;

   m = CBuf->marks;
   if (m == NULL) return(0);

   if (go) jed_goto_mark(m);
   if (m->flags & VISIBLE_MARK_MASK)
     {
        CBuf->vis_marks--;
        /* touch screen since region may be highlighted */
        if (CBuf->vis_marks == 0) touch_screen();
     }

   deallocate_mark (&CBuf->mark_array, &CBuf->marks);

   return 1;
}



int mark_spot()
{
    push_spot();
    message("Spot Marked.");
    return(1);
}

static int pop_spot_go (int go)
{
   Mark *m;

   m = CBuf->spots;
   if (m == NULL) return(0);

   if (go) jed_goto_mark (m);

   deallocate_mark (&CBuf->spot_array, &CBuf->spots);
   return 1;
}

int pop_spot ()
{
   return pop_spot_go (1);
}

int exchange_point_mark(void)
{
   Line *save_line;
   int save_point;
   unsigned int save_n;
   Mark *m;

   if ((m = CBuf->marks) == NULL) return(0);

   save_point = Point;
   save_line = CLine;
   save_n = LineNum + CBuf->nup;

   jed_goto_mark (m);

   m->point = save_point; m->line = save_line; m->n = save_n;
   return(1);
}

////
int rectangle_to_lineblock(void)
{
    // convert rectangle to lineblock
    // mark will be at eol on north line
    // limiting point will be at eol south line

    int l1, l2;

    // no rectangle, nothing to do
    if (!check_rectangle()) return 0;

    // rectangle coordinates are in globale variables

    //  current line
    l1 = LineNum;

    // line holding block mark
    l2 = CBuf->marks->n;

    //  cursor to be positioned after mark

    //  change block type
    CBuf->blocktype=0;

    // current line north
    if (l1 < l2)
    {
        Point=0;
        exchange_point_mark();
        eol();
    }

    // current line south
    else if (l1 > l2)
    {
        push_spot();
        goto_line (&l2);
        Point=0;
        exchange_point_mark();
        pop_spot();
        eol();
    }

    else if (l1==l2)
    {
        Point=0;
        exchange_point_mark();
        eol();
    }

    return 1;
}

////
int stream_to_lineblock(void)
// convert char stream to lineblock
// mark will be at eol on north line
// limiting point will be at eol south line
{
    int l1, l2;

    // no mark or wrong blocktype
    if (CBuf->marks == NULL || CBuf->blocktype!=0)
        return 0;

    //  current line
    l1   = LineNum;

    // line holding block mark
    l2   = CBuf->marks->n;

    //  cursor to be positioned after mark

    // current line north
    if (l1 < l2)
    {
        Point=0;
        exchange_point_mark();
        eol();
    }

    // current line south
    else if (l1 > l2)
    {
        push_spot();
        goto_line (&l2);
        Point=0;
        exchange_point_mark();
        pop_spot();
        eol();
    }

    else if (l1==l2)
    {
        Point=0;
        exchange_point_mark();
        eol();
    }

    //  change block type
    CBuf->blocktype=2;

    RestoreHilight=2;

    return 1;
}

////
int lineblock_to_stream(void)
// convert line block to char stream in canonical form
{
    int l1, l2;

    // no mark or wrong blocktype
    if (CBuf->marks == NULL || CBuf->blocktype!=2)
        return 0;

    //  current line
    l1 = LineNum;

    // line holding block mark
    l2 = CBuf->marks->n;

    //  change block type
    CBuf->blocktype=0;

    //  position cursor after mark and enclose everything
    if (l1 < l2)
    {
        bol();
        exchange_point_mark();
        eol();
    }
    else eol();

    // has canonical form

    return 1;
}

////
int check_stream (void)
// determine character block coordinates
// without touching mark, returning:
// -1   error
//  0   no mark, no char mark (rectangle)
//  number of lines
{
    Line *beg, *end;
    int l1, l2;

    // too many visible marks (shouldn't happen)
    if (CBuf->vis_marks > 1)
    {
        CBuf->vis_marks = 0;
        CBuf->marks = NULL;
        jed_verror ("More than one visible mark!");
        return -1;
    }

    // no mark or wrong blocktype
    if (CBuf->marks == NULL || CBuf->blocktype==1)
    {
        BegPoint    = 0;
        EndPoint    = 0;
        BegRow      = 0;
        EndRow      = 0;
        return 0;
    }

    //  tell char stream from whole lines
    if (CBuf->blocktype==2) RestoreHilight=2;
    else RestoreHilight=0;

    //  current delimiting line
    end = CLine;
    EndPoint = Point;
    l1 = LineNum;

    // delimiting line holding block mark
    beg = CBuf->marks->line;
    BegPoint = CBuf->marks->point;
    l2 = CBuf->marks->n;

    if (l1==l2)
    {
        // stream on one line only
        BegL = CLine;
        EndL = CLine;
        BegRow   = LineNum;
        EndRow   = LineNum;
        if (BegPoint > EndPoint)
        {
            BegPoint = Point;
            EndPoint = CBuf->marks->point;
        }
        return 1;
    }

    // stream on more than one line

    // point earlier than mark
    if (l1 < l2)
    {
        BegL       = end;
        BegRow     = l1;
        BegPoint   = Point;
        EndL       = beg;
        EndRow     = l2;
        EndPoint   = BegPoint;
    }

    //  mark earlier than point
    else
    {
        BegL       = beg;
        BegRow     = l2;
        //BegPoint
        EndL       = end;
        EndRow     = l1;
        //EndPoint
    }

    return (1 + EndRow - BegRow);
    // BegL, EndL, BegRow, EndRow, BegPoint, EndPoint
    // are now correctly determined and globally available
}

/*{{{ Narrow/Widen/Region Functions */

 /*returns 0 if the mark is not set and gives error.  Exchanges point and mark
  * to produce valid region.  A valid region is one with mark
  * earlier in the buffer than point.  Always call this if using a region
  * which reqires point > mark.  Also, push spot first then pop at end.
  */
int check_region0(int *push)
{
   register Line *beg, *tthis = CLine;
   int pbeg;

   if (CBuf->marks == NULL)
     {
        msg_error("No region defined");
        return(0);
     }

   if (*push) push_spot();
   beg = CBuf->marks->line;
   pbeg = CBuf->marks->point;

   if (beg == CLine)
     {
        if (pbeg <= Point) return(1);
     }

   else
     {
        while((beg != NULL) && (beg != tthis)) beg = beg->next;
        if (beg == tthis) return(1);
     }

   exchange_point_mark();
   return(1);
}

////
int check_region (int * push)
//  Returns 0 if no mark set or if mark is for rectangle.
//  Will handle line block converting it to char stream.
//  Will exchange point and mark to produce canonical region,
//  i.e. a region with mark earlier than point.
//  Can push spot to be popped when done (after returning).
{
    int ln_cline, ln_mline, point_mline;

    // too many visible marks (shouldn't happen)
    if (CBuf->vis_marks > 1)
    {
        CBuf->vis_marks = 0;
        CBuf->marks = NULL;
        jed_verror ("More than one visible mark!");
        return 0;
    }

    // no mark, wrong blocktype
    if (CBuf->marks == NULL || CBuf->blocktype==1)
        return 0;

    if (*push) push_spot();
    // If pushed, it will not be popped here!
    // pop will be after returning!

    if (CBuf->blocktype==2)
        return lineblock_to_stream();

    // line block is converted to char stream
    // and is already in canonical form

    //  current line
    ln_cline = LineNum;

    // line holding block mark
    ln_mline    = CBuf->marks->n;
    point_mline = CBuf->marks->point;

    // block extends on one line only
    if (ln_cline==ln_mline)
    {
        if (point_mline > Point) exchange_point_mark();
    }

    //  mark south, current line north
    else if (ln_mline > ln_cline) exchange_point_mark();

    // char stream in canonical form

    return 1;
    // remember to pop spot if spot was pushed
}

////
int restore_selection (void)
//  revive highlight of last selection
//  only meaningful immediately after cancelling it
{
    int BegCol, EndCol, ok=0;
    int lnpnt=Point;

    // do nothing if read-only
    if (CBuf->flags & READ_ONLY) return 0;

    // do nothing if there is a mark
    if (CBuf->marks != NULL) return 0;

    if (RestoreHilight==1) return restore_rectangle();

    // coordinates of stream must be known
    // for char stream, cursor must be at one end

    if (RestoreHilight==0)
    {
        if ((unsigned) BegRow == LineNum && BegPoint == Point) ok=2;
        else if ((unsigned) EndRow == LineNum && EndPoint == Point) ok=1;
    }

    else if(RestoreHilight==2)
    {
        if (BegRow == EndRow) ok=1;
        else if ((unsigned) BegRow == LineNum) ok=2;
        else ok=1;
    }

    BegCol = BegPoint + 1;
    EndCol = EndPoint + 1;

    if (ok < 1) return 0;
    // cursor has moved from corners, restoring may be messy

    else if (ok==1)
    {
        //  canonical case
        goto_line (&BegRow);
        goto_column (&BegCol);

        if(RestoreHilight==2) toggle_block (&Number_Two);
        else set_mark_cmd();

        goto_line (&EndRow);
        goto_column (&EndCol);
    }

    else if (ok==2)
    {
        //  anti-canonical
        goto_line (&EndRow);
        goto_column (&EndCol);

        if(RestoreHilight==2) toggle_block (&Number_Two);
        else set_mark_cmd();

        goto_line (&BegRow);
        goto_column (&BegCol);
    }

    if(RestoreHilight==2) Point=lnpnt;
    //  works only downwards
    //  because of the way streams are highlighted
    //  so if the full lines block was canonical
    //  cursor stays put otherwise it will move to bol

    // selection is highlighted
    return 1;
}

////static int widen_buffer_lines (Buffer *b)
////int widen_buffer (Buffer *b)
////int narrow_to_region (void)
////int widen (void)
////int widen_region (void)
////int narrow_to_lines ()

//// custom
int yank()
{
    CHECK_READ_ONLY

    // nothing to paste
    if (Paste_Buffer==NULL && Rectangle_Buffer==NULL)
        return 0;

    if (CBuf->vis_marks) return 0;
    // no insertion if there is a block

    // paste rectangle
    if (clipBoard == 1)
    {
        if (IN_MINI_WINDOW) return 0;
        //  don't insert rectangle in minibuffer

        int r = insert_rectangle();

        //  if insertion failed, stop here
        if (!r) return 0;

        //  set flag for restoring selection
        RestoreHilight = 1;

        return 1;
        //  rectangle was successfully inserted
    }

    // paste stream
    if (clipBoard == 0)
    {
        int tmp = RestoreHilight;

        if (!IN_MINI_WINDOW && RestoreHilight == 2)
        {
            eol();
            jed_trim_whitespace ();
            if (calculate_column() > 1)
                jed_insert_newline();
            //  so the lineblock inserted is neatly separated
        }

        insert_buffer (Paste_Buffer);
        RestoreHilight = tmp;
        return 1;
    }

    // should not happen
    jed_verror ("Unable to select buffer to paste from.");
    return 0;
}

int copy_region_to_buffer0(Buffer *b)
{
   int first_point, last_point, n;
   Line *first, *last;
   Buffer *save_buf;

   if (b->flags & READ_ONLY)
     {
        msg_error(Read_Only_Error);
        return (0);
     }

   if (!check_region(&Number_One)) return(0);  /* spot pushed */
   last = CLine;
   last_point = Point;

   jed_pop_mark(1);
   if (b == CBuf)
     {
        msg_error("A buffer cannot be inserted upon itself.");
        pop_spot();
        return(0);
     }

   first = CLine;
   first_point = Point;

   save_buf = CBuf;
   switch_to_buffer(b);

   /* go through standard routines for undo compatibility */
   Suspend_Screen_Update = 1;
   if (first == last)
     {
        n = last_point - first_point;
        if (save_buf == MiniBuffer)
          {
             (void) jed_insert_nbytes (first->data + first_point, n);
          }
        else (void) jed_quick_insert (first->data + first_point, n);
     }
   else
     {
        n = first->len - first_point;
        if (-1 == jed_quick_insert(first->data + first_point, n))
          goto the_return;

        while (first = first->next, first != last)
          {
             if (-1 == jed_quick_insert (first->data, first->len))
               goto the_return;
          }
        (void) jed_quick_insert(first->data, last_point);
     }

   the_return:
   switch_to_buffer(save_buf);
   pop_spot();
   return(1);
}

int copy_to_pastebuffer0()
{
   /* delete paste buffer */
   if (Paste_Buffer != NULL) delete_buffer(Paste_Buffer);
   Paste_Buffer = make_buffer (" <paste>", NULL, NULL);

   copy_region_to_buffer(Paste_Buffer);
   return(0);
}

////
int copy_region_to_buffer (Buffer *b)
{
    int first_point, last_point, n, fulln=0;
    Line *first, *last;
    Buffer *save_buffer;

    if (b->flags & READ_ONLY)
    {
        flush_message (Read_Only_Error);
        return 0;
    }

    if (b == CBuf)
    {
        msg_error ("A buffer cannot be inserted into itself.");
        return 0;
    }

    // no rectangles here
    if (CBuf->blocktype==1) return 0;

    //  if whole lines, take note
    if (CBuf->blocktype==2) fulln=1;

    if (!check_region (&Number_One)) return 0;
    // spot pushed

    last = CLine;
    last_point = Point;

    jed_pop_mark (1);

    first = CLine;
    first_point = Point;

    save_buffer = CBuf;

    switch_to_buffer (b);

    /* go through standard routines for undo compatibility */
    Suspend_Screen_Update = 1;

    if (first == last)
    {
        // block is on one line only
        n = last_point - first_point;

        if (save_buffer == MiniBuffer)
            (void) jed_insert_nbytes (first->data + first_point, n);

        else (void) jed_quick_insert (first->data + first_point, n);
    }

    else
    {
        // block extends on more than one line

        // copy portion on first row

        n = first->len - first_point;

        if (-1 == jed_quick_insert (first->data + first_point, n))
            goto the_return;

        // copy block lines in between
        while (first = first->next, first != last)
        {
            if (-1 == jed_quick_insert (first->data, first->len))
                goto the_return;
        }

        // copy portion on last row

        (void) jed_quick_insert (first->data, last_point);
    }

    the_return:

    if (fulln) RestoreHilight=2;
    else RestoreHilight=1;
    //  set correct restore flag

    switch_to_buffer (save_buffer);

    pop_spot();

    if (b==Paste_Buffer) clipBoard=0;;

    //restore_selection();
    //  selection does not need to be restored after copying

    return 1;
}

////
int copy_to_pastebuffer()
{
    // nothing to copy if block is rectangle
    if (check_rectangle()) return 0;

    // delete paste buffer, create new empty one
    if (Paste_Buffer != NULL) delete_buffer (Paste_Buffer);
    Paste_Buffer = make_buffer (" <paste>", NULL, NULL);

    copy_region_to_buffer (Paste_Buffer);

    // indicate which of paste/rectangle buffer
    // will be used for next paste
    clipBoard=0;

    return 0;
}

////
void DeletionCheck(void)
// only to be called from delete_region
{
    int tmp=0, action=0;
    static int this_ln, this_pn;

    if (DeletionToKillArray==0)
    {
        // this has been set in kill_region
        // to avoid cutting region to clipboard
        // while writing it to kill_array

        // reverse the flag, do nothing else
        DeletionToKillArray=1;
        return;
    }

    // normal case, deletion goes to kill_array
    // pastebuffer and kill_array keep separate

    // region is already in canonical form

    check_stream();

    // append or copy to kill_array?
    // append when repeatedly deleting words on same line
    // without moving the cursor
    // inspired by wordperfect/dos

    // append  or copy?
    // action 0: copy, action 1: append
    action =
    (thisbuf==CBuf && this_ln==BegRow && this_pn==BegPoint);

    thisbuf = CBuf;
    this_ln = BegRow;
    this_pn = BegPoint;

    dup_mark();
    // required because writing to kill_array destroys mark

    if (action) append_region_to_kill_array (&tmp);
    else copy_region_to_kill_array (&tmp);
}

static int jed_check_readonly_region (void)
{
   int beg_point, end_point;
   Line *beg, *end;

   if (CBuf->flags & READ_ONLY)
     {
        msg_error(Read_Only_Error);
        return -1;
     }

   if (!check_region(&Number_Zero))
     return -1;

   end = CLine; end_point = Point;
   (void) exchange_point_mark ();
   beg = CLine; beg_point = Point;
   (void) exchange_point_mark ();

   while (1)
     {
        if (beg->flags & JED_LINE_IS_READONLY)
          {
             if ((beg == end)
                 && (end_point == 0)
                 && (beg_point == 0))
               return 0;

             msg_error (Line_Read_Only_Error);
             return -1;
          }
        if (beg == end)
          break;

        beg = beg->next;
     }
   return 0;
}

int delete_region (void)
{
   int beg_point, end_point;
   Line *beg, *end;

   if (0 != jed_check_readonly_region ())
     return -1;

    DeletionCheck();
    ////  save or not save deletion to kill_array
    ////  saving actually comes before deletion

   /* make this go through standard ins/del routines to ensure undo */

   end = CLine; end_point = Point;
   push_spot();
   jed_pop_mark(1);
   beg = CLine; beg_point = Point;
   pop_spot();

   if (end != beg)
     {
        bol ();

        if (-1 == jed_del_nbytes (end_point))
          return -1;

        /* go back because we do not want to mess with Line structures
           changing on us --- shouldn't happen anyway */

        while (jed_up (1) && (CLine != beg))
          {
             bol ();
             if (-1 == jed_del_through_eol ())
               return -1;
          }
        end_point = CLine->len;                /* include \n */
     }

   jed_set_point (beg_point);
   (void) jed_generic_del_nbytes (end_point - Point);
   return 1;
}

int kill_region0()
{
   if (-1 == jed_check_readonly_region ())
     return -1;

   /* need two marks for this one */
   push_spot();
   if (!jed_pop_mark(1))
     {
        check_region(&Number_Zero);
        pop_spot();
        return(0);
     }
   jed_push_mark();
   jed_push_mark();
   pop_spot();

   copy_to_pastebuffer();
   delete_region();
   return(1);
}

////
int kill_region()
//  it affects the pastebuffer, not the kill_array
{
    int tmpm = 1;

    if (-1 == jed_check_readonly_region ())
        return -1;
    CHECK_READ_ONLY

    if (CBuf->blocktype==2) check_region (&Number_Zero);
    // check_region is called again
    // but it had to be done unconditionally
    // to convert the line block to stream

    /* need two marks for this one */
    push_spot();

    if (!jed_pop_mark(1))
    {
        check_region (&Number_Zero);
        pop_spot();
        return 0;
    }
    jed_push_mark();
    jed_push_mark();
    pop_spot();

    copy_to_pastebuffer();

    // since already in pastebuffer, skip kill_array
    DeletionToKillArray=0;
    delete_region();
    //DeletionToKillArray=1;

    return 1;
}

////
int visible_mark (void)
//  if there is no mark return 0
//  if there is no visible mark return 0
//  otherwise return 1
{
    if
    (
        (CBuf->marks == NULL) ||
        (0 == CBuf->marks->flags & VISIBLE_MARK)
    )
        return 0;

    return 1;
}

////
//  blktype 1
//      if rectangle, cancel
//      if other block, cancel, start rectangle
//      if no block, start rectangle
//  blktype 2
//      if line, cancel
//      if other block, cancel, start line
//      if no block, start line
//  blktype 0
//      cancel any block, set blocktype to 0

int toggle_block (int *blktype)
{
    if (CBuf->vis_marks > 1)
    {
        CBuf->marks=NULL;
        CBuf->vis_marks=0;
        jed_verror ("Error: more than one visible mark!");
        return 0;
    }

    if (IN_MINI_WINDOW) *blktype = 0;

    // column block
    if (*blktype==1)
    {
        // no visible marks, start column block
        if (!visible_mark())
        {
            set_mark_cmd ();
            CBuf->blocktype = 1;

            return 1;
        }

        // there is a column block, cancel it, return
        if (visible_mark() && CBuf->blocktype==1)
        {
            jed_pop_mark(0);
            CBuf->blocktype=0;
            return 1;
        }

        // there is a non-column block, cancel it, start column block
        if (visible_mark() && CBuf->blocktype!=1)
        {
            jed_pop_mark(0);
            set_mark_cmd ();
            CBuf->blocktype = 1;
            update_sans_update_hook_cmd (&Number_One);
            // refresh the start mark column

            return 1;
        }
    }

    // line block, entire line selected wherever the cursor
    else if (*blktype==2)
    {
        // no visible marks, start line block
        if (!visible_mark())
        {
            push_spot();
            bol();
            set_mark_cmd();
            CBuf->blocktype = 2;
            pop_spot();
            return 1;
        }

        // there is a line block, cancel it
        if (visible_mark() && CBuf->blocktype==2)
        {
            jed_pop_mark(0);
            CBuf->blocktype=0;
            return 1;
        }

        // there is a non-line block, cancel it, start line block
        if (visible_mark() && CBuf->blocktype!=2)
        {
            push_spot();
            bol();
            jed_pop_mark(0);
            set_mark_cmd();
            CBuf->blocktype = 2;
            pop_spot();
            return 1;
        }
    }

    // character block (stream)
    // will set blocktype to default
    // will cancel visible mark
    else if (*blktype==0)
    {
        // cancel last mark
        if (visible_mark()) jed_pop_mark(0);

        CBuf->blocktype = 0;
        return 1;
    }

    // better safe than sorry
    jed_verror ("Unknown block error!");
    return 0;
}

////
int block_type (void)
{
    return CBuf->blocktype;
}

////
int check_rectangle(void)
// determines the North/West/South/East borders of rectangle
{
    int c1, c2, l1, l2, lmin, lmax, cmin, cmax;

    // nothing to do if no rectangle
    if (CBuf->vis_marks != 1) return 0;
    if (CBuf->blocktype != 1) return 0;

    // current position on rectangle corner
    // cursor is always inside rectangle
    c1 = calculate_column();
    l1 = CBuf->linenum;
    l2 = CBuf->marks->n;

    exchange_point_mark();
    c2 = calculate_column();
    exchange_point_mark();

    if (c2>c1)
    {
        cmax = c2;
        cmin = c1;
    }
    else
    {
        cmax = c1;
        cmin = c2;
    }

    if (l2>l1)
    {
        lmax = l2;
        lmin = l1;
    }
    else
    {
        lmax = l1;
        lmin = l2;
    }

    int tmp_c = (c1 == cmax);
    // 0    cursor on left  border
    // 1    cursor on right border

    int tmp_r = (l1 == lmax);
    // 0    cursor on top    border
    // 1    cursor on bottom border

    N_ROW = lmin;
    W_COL = cmin;
    S_ROW = lmax;
    E_COL = cmax;

    PULSE_C = tmp_c;
    PULSE_R = tmp_r;

    return 1;
}

////
int shift_linetext_right(void)
//  shift text one cell to the right, nothing is deleted
{
    int lmin, lmax;
    int tmpnt = Point;

    CHECK_READ_ONLY

    if (CBuf->blocktype != 2) return 0;

    if (check_stream() < 1) return 0;

    // lines checked, borders known
    lmin = BegRow;
    lmax = EndRow;

    Suspend_Screen_Update = 1;

    //  if point after mark, reverse
    if (lmax==LineNum)
    {
        check_region(&Number_Zero);
        exchange_point_mark();
        //  side effect, lineblock converted to stream
    }

    // cursor at top of block

    while (1)
    {
        bol();
        if (-1 == _jed_ins_byte (' ')) break;
        if (LineNum==(unsigned) lmax) break;
        CLine = CLine->next;
        LineNum++;
    }

    goto_line (&lmin);

    stream_to_lineblock();

    Suspend_Screen_Update = 0;
    Point = ++tmpnt;
    return 1;
}

////
int shift_linetext_left(void)
//  shift text one column left in line
//  deleting first char cell if whitespace
{
    int  lmin, lmax;
    int tmpnt = Point;

    CHECK_READ_ONLY

    if (CBuf->blocktype !=2) return 0;

    if (check_stream() < 1) return 0;

    // lines checked, borders known
    lmin = BegRow;
    lmax = EndRow;

    Suspend_Screen_Update = 1;

    //  if point after mark, reverse
    if (lmax==LineNum)
    {
        check_region(&Number_Zero);
        exchange_point_mark();
        //  side effect, lineblock converted to stream
    }

    // cursor at top of block

    while (1)
    {
        bol();
        //  delete whitespace only
        if ((int) *(CLine->data) == 32 || (int) *(CLine->data) == 9)
        {
            if (-1 == jed_del_nbytes (1)) break;
        }

        if (LineNum==(unsigned) lmax) break;
        CLine = CLine->next;
        LineNum++;
    }

    goto_line (&lmin);
    stream_to_lineblock();
    if (tmpnt > 0) Point = --tmpnt;
    else Point=0;

    Suspend_Screen_Update = 0;

    return 1;
}

////
void draw_rectangle (int north_l, int west_c, int south_l, int east_c)
{
    int check_coordinates =
        (north_l > 0)        +
        (west_c  > 0)        +
        (south_l >= north_l) +
        (east_c  >= west_c)  ;

    if(check_coordinates < 4)  return;
    if (CBuf->marks != NULL) return;

    goto_line    (&north_l);
    goto_column  (&west_c);

    toggle_block (&Number_One);

    goto_line    (&south_l);
    goto_column  (&east_c);
}

////
int restore_rectangle (void)
//  revive highlight of last rectangle
{
    int lmin, lmax, cmin, cmax, cursor_c, cursor_r;

    CHECK_READ_ONLY
    if (CBuf->marks != NULL) return;

    // rectangle coordinates are known

    if (!N_ROW) return 0;
    // nothing to restore

    if(PULSE_R) goto_line(&S_ROW);
    else goto_line(&N_ROW);

    if(PULSE_C) goto_column(&E_COL);
    else goto_column(&W_COL);

    lmin = N_ROW;
    lmax = S_ROW;
    cmin = W_COL;
    cmax = E_COL;
    cursor_c = PULSE_C;
    cursor_r = PULSE_R;

    // the pairs (0,0), (0,1), (1,0), (1,1) identify corners

    int height, i, tmp, corner;

    corner = 10*cursor_r + cursor_c;
    //  0 NW, 1 NE, 10 SW, 11 SE

    height = S_ROW - N_ROW;
    tmp = E_COL+1;

    if (corner == 0)
    {
        // cursor is NW, go SE

        goto_line (&S_ROW);
        goto_column (&E_COL);

        // start rectangle and re-init variables in screen.c
        toggle_block (&Number_One);

        bol();
        for (i=0; i<height; i++)
        {
            jed_up(1);
            goto_column (&tmp);
        }

        goto_column (&W_COL);
        // takes care of nil height as well
    }

    else if (corner == 1)
    {
        // cursor is NE, go SW

        goto_line (&S_ROW);
        goto_column (&W_COL);

        // start rectangle and re-init variables in screen.c
        toggle_block (&Number_One);

        // highlight rectangle without sudden jumps
        for (i=0; i<height; i++)
        {
            goto_column (&tmp);
            jed_up(1);
        }

        goto_column (&E_COL);
        // takes care of nil height as well
    }

    else if (corner == 10)
    {
        // cursor is SW, go NE

        goto_line (&N_ROW);
        goto_column (&E_COL);

        // start rectangle and re-init variables in screen.c
        toggle_block (&Number_One);

        jed_down (height);
        goto_column (&W_COL);
        // takes care of nil height as well
    }

    else if (corner == 11)
    {
        // cursor is SE, go NW

        goto_line (&N_ROW);
        goto_column (&W_COL);

        // start rectangle and re-init variables in screen.c
        toggle_block (&Number_One);

        // highlight rectangle without sudden jumps
        for (i=0; i<height; i++)
        {
            goto_column (&tmp);
            jed_down (1);
        }

        goto_column (&E_COL);
        // takes care of nil height as well
    }

    else
    {
        flush_message("Cannot restore rectangle.");
        return 0;
    }

    // rectangle is highlighted, reset coordinates

    N_ROW   = 0;
    S_ROW   = 0;
    W_COL   = 0;
    E_COL   = 0;
    PULSE_C = 0;
    PULSE_R = 0;

    return 1;
}

////
int insert_rectangle()
//  return 1 for success, 0 for failure
{
    int c1, c2;
    Line *rline;

    CHECK_READ_ONLY

    if (Rectangle_Buffer == NULL) return 0;
    rline =  Rectangle_Buffer->beg;

    if (rline == NULL) return 0;

    if (CBuf->marks != NULL) return 0;
    // no insertion if there is a block

    // NW
    c1 = calculate_column();

    while (1)
    {
        goto_column (&c1);

        if (-1 == jed_quick_insert (rline->data, rline->len))
            return 0;

        rline = rline->next;
        if (rline == NULL) break;

        if (0 == jed_down (1))
        {
            eol();
            jed_insert_newline();
        }
    }

    // cursor is just right of SE corner

    c2 = calculate_column();
    if (c2 > 1) jed_left(1);

    return 1;
}

////
int shift_rectangle_right(void)
//  shift entire rectangle one column to right side
{
    int lmin, lmax, cmin, cmax;

    CHECK_READ_ONLY

    if (!check_rectangle()) return 0;
    toggle_block (&Number_One);
    // block cancelled

    // rectangle checked, we know its borders
    // increase columns by 1 to account for shift
    lmin    = N_ROW;
    lmax    = S_ROW;
    cmin    = W_COL++;
    cmax    = E_COL++;

    goto_line (&lmin);

    while (1)
    {
        goto_column (&cmin);
        if (-1 == _jed_ins_byte ((unsigned char) ' ')) break;
        if (LineNum==(unsigned) lmax) break;
        CLine = CLine->next;
        LineNum++;
    }

    restore_rectangle();
    // shifted rectangle is highlighted

    return 1;
}

////
int shift_rectangle_left(void)
//  shift entire rectangle one column to left side
//  stop at begin of line
{
    int lmin, lmax, cmin, cmax;

    CHECK_READ_ONLY

    if (!check_rectangle()) return 0;
    toggle_block (&Number_One);
    // block cancelled

    // rectangle checked, we know its borders
    // decrease columns by 1 to account for shift

    lmin    = N_ROW;
    lmax    = S_ROW;
    cmin    = W_COL--;
    cmax    = E_COL--;

    if (cmin==1) return 0;
    // rectangle is at bol, no shifting

    goto_line (&lmin);

    while (1)
    {
        int cc=cmin-1;
        goto_column (&cc);
        if (-1 == jed_del_nbytes (1)) break;
        if (LineNum==(unsigned) lmax) break;
        CLine = CLine->next;
        LineNum++;
    }

    restore_rectangle();
    // shifted rectangle is highlighted

    return 1;
}

////
int blank_or_open_rectangle (int overwrite)
//  to open a rectangle means to fill the left column with spaces
//  while shifting its content right
//  this will be done when argument 1 passed
//  for argument 0, rectangle content will be wiped out
{
    int  lmin, lmax, cmin, cmax, w;

    CHECK_READ_ONLY

    if (!check_rectangle()) return 0;

    // rectangle checked, we know its borders
    lmin = N_ROW;
    cmin = W_COL;
    lmax = S_ROW;
    cmax = E_COL;

    w = cmax - cmin + 1;

    // The width w corresponds to the new rectangles
    // with cursor always inside, that's the reason for the +1.

    Suspend_Screen_Update = 1;
    //jed_pop_mark (1);
    //CBuf->blocktype = 0;
    toggle_block(&Number_One);
    goto_line (&lmin);
    goto_column (&cmin);
    push_spot();

    // push_spot needed to return to NW corner when done

    while (1)
    {
        goto_column (&cmin);

        if (overwrite)
            if (-1 == jed_del_nbytes (w)) break;

        //if (-1 == jed_ins_char_n_times (' ', w)) break;
        if (-1 == jed_insert_wchar_n_times(' ', w)) break;
            // will fail if more than 20*JED_MAX_MULTIBYTE_SIZE];

        if (LineNum==(unsigned) lmax) break;
        //CLine = CLine->next;
        //LineNum++;
        (void) jed_down (1);
    }

    pop_spot();
    // position cursor on NW corner of cancelled rectangle

    Suspend_Screen_Update = 0;
    update_sans_update_hook_cmd (&Number_One);
    return 1;
}

////
int open_rectangle (void)
{
    // fill rectangle with spaces
    // while shifting its content right
    return blank_or_open_rectangle (0);
}

////
int blank_rectangle (void)
{
    // fill rectangle with spaces
    // overwriting its content
    return blank_or_open_rectangle (1);
}
////

/*{{{ Rectangle Functions */

static char *Rect_Error = "Rectangle has 0 width.";
int insert_rectangle0()
{
   int c1;
   Line *rline;

   CHECK_READ_ONLY
   if (0 == buffer_exists (Rectangle_Buffer))
     {
        Rectangle_Buffer = NULL;
        return 0;
     }

   Suspend_Screen_Update = 1;
   c1 = calculate_column();
   rline = Rectangle_Buffer->beg;
   if (rline != NULL) while (1)
     {
        goto_column(&c1);
        if (-1 == jed_quick_insert (rline->data, rline->len))
          return -1;

        rline = rline->next;
        if (rline == NULL) break;
        if (0 == jed_down (1))
          {
             eol();
             jed_insert_newline();
          }
     }
   return(1);
}

int open_rectangle0()
{
   int c1, n, c2, tmpm;
   Line *save_line;

   CHECK_READ_ONLY
   if (!check_region(&Number_One)) return(0); /* push_spot(); performed */

   c1 = calculate_column();
   save_line = CLine;
   tmpm = 1; jed_pop_mark(1);
   c2 = calculate_column();
   n = c2 - c1;
   if (n < 0)
     {
        n = -n;
        c1 = c2;
     }

   Suspend_Screen_Update = 1;
   while(1)
     {
        goto_column(&c1);
        if (-1 == jed_insert_wchar_n_times(' ', n))
          break;
        if (CLine == save_line) break;
        (void) jed_down (1);
     }
   pop_spot();

   return(1);
}

/* MB Safe */
////static int copy_or_kill_rectangle (int kill)
int copy_or_kill_rectangle (int kill)
{
    Line *save_line, *line, *beg;
    int c1, c2, dc, tmp;

    if (!check_region(&Number_One)) return(0);
    /* spot pushed */
    /* delete Rectangle buffer */
    if (Rectangle_Buffer != NULL) delete_buffer(Rectangle_Buffer);

    Rectangle_Buffer = make_buffer (" <rect>", NULL, NULL);
    c2 = calculate_column();
    save_line = CLine;

    jed_pop_mark(1);
    c1 = calculate_column();
    if (c1 == c2)
    {
        msg_error(Rect_Error);
        pop_spot();
        return(0);
    }
    if (c1 > c2)
    {
        tmp = c1;
        c1 = c2;
        c2 = tmp;
        goto_column(&c1);
    }

    dc = c2 - c1;

    /* go through the region copying rectanglar blocks to Rectanglebuffer */

    line = beg = NULL;
    while (1)
    {
        int col2, col1;
        unsigned int len, len1;
        unsigned int nspaces;
        unsigned int nbytes;
        unsigned char *p1, *p2;

        col1 = goto_column1 (&c1);
        p1 = CLine->data + Point;

        if (col1 == c1)
        {
            col2 = goto_column1(&c2);
            p2 = CLine->data + Point;
        }
        else
        {
            col2 = col1;
            p2 = p1;
        }

        nspaces = (unsigned int) (dc - (col2 - col1));
        nbytes = (unsigned int) (p2 - p1);
        len1 = len = nbytes + nspaces;

        /* Need to allocate at least 2 bytes, since a single byte is a
         * signature of a line with a single character whose value is a
         * newline character.  See make_line1 for details.
         */
        if (len <= 1) len1++;

        if (beg == NULL)
        {
             beg = line = make_line1 (len1);
             beg->prev = NULL;
        }
        else
        {
             line->next = make_line1 (len1);
             line->next->prev = line;
             line = line->next;
        }
        line->len = len;

        memcpy ((char *) line->data, (char *) p1, nbytes);
        memset ((char *) line->data + nbytes, ' ', nspaces);

        if (kill && (nbytes != 0))
        {
            jed_position_point (p1);
            if (-1 == jed_del_nbytes (nbytes))
                break;
        }

         if (CLine == save_line) break;
         (void) jed_down(1);
    }

     line->next = NULL;

     Rectangle_Buffer->line = Rectangle_Buffer->beg = beg;
     Rectangle_Buffer->end = line;
     Rectangle_Buffer->point = 0;

     pop_spot();
     return 0;
}

int copy_rectangle0 (void)
{
   return copy_or_kill_rectangle (0);
}

int kill_rectangle0 (void)
{
   return copy_or_kill_rectangle (1);
}

int blank_rectangle0 (void)
{
   int c1, c2;
   Line *save_line;
   int ncols;

   CHECK_READ_ONLY
   if (!check_region(&Number_One)) return(0); /* push_spot(); performed */

   c1 = calculate_column();
   save_line = CLine;
   jed_pop_mark(1);
   c2 = calculate_column();
   if (c1 > c2)
     {
        int tmp = c1;
        c1 = c2;
        c2 = tmp;
     }
   ncols = c2 - c1;

   Suspend_Screen_Update = 1;
   while(1)
     {
        int pnt;

        goto_column (&c2);
        pnt = Point;
        goto_column (&c1);

        if (-1 == jed_del_nbytes (pnt - Point))
          break;

        if (-1 == jed_insert_wchar_n_times( ' ', ncols))
          break;

        if (CLine == save_line) break;
        (void) jed_down (1);
     }
   pop_spot();
   return(1);
}

////
int copy_or_cut_rectangle (int kill)
{
    Line *save_line, *line, *beg;
    int c1, c2, dc, tmp;

    push_spot();
    if (!check_rectangle())
    {
        pop_spot();
        return 1;
    }
    // rectangle checked, we know its borders
    toggle_block (&Number_One);

    c1 = W_COL;
    c2 = E_COL;

    //  our rectangles include cursor
    //  correction for loop below
    c2++;

    goto_line (&S_ROW);
    save_line = CLine;
    goto_line (&N_ROW);

    // delete Rectangle buffer
    if (Rectangle_Buffer != NULL)
        delete_buffer(Rectangle_Buffer);

    Rectangle_Buffer = make_buffer (" <rect>", NULL, NULL);
    clipBoard=1;

    dc = c2 - c1;

    //  copy rectangle segments
    //  going down from north-west corner

    line = NULL;
    beg  = NULL;

    while (1)
    {
        int col2, col1;
        unsigned int len, len1;
        unsigned int nspaces;
        unsigned int nbytes;
        unsigned char *p1, *p2;

        col1 = goto_column1 (&c1);
        p1 = CLine->data + Point;

        if (col1 == c1)
        {
            col2 = goto_column1 (&c2);
            p2 = CLine->data + Point;
        }
        else
        {
            col2 = col1;
            p2 = p1;
        }

        nspaces = (unsigned int) (dc - (col2 - col1));
        nbytes = (unsigned int) (p2 - p1);
        len1 = nbytes + nspaces;
        len  = nbytes + nspaces;

        /* Need to allocate at least 2 bytes, since a single byte is a
         * signature of a line with a single character whose value is a
         * newline character.  See make_line1 for details.
         */
        if (len <= 1) len1++;

        if (beg == NULL)
        {
             beg = line = make_line1 (len1);
             beg->prev = NULL;
        }
        else
        {
             line->next = make_line1 (len1);
             line->next->prev = line;
             line = line->next;
        }
        line->len = len;

        memcpy ((char *) line->data, (char *) p1, nbytes);
        memset ((char *) line->data + nbytes, ' ', nspaces);

        if (kill && (nbytes != 0))
        {
            jed_position_point (p1);
            if (-1 == jed_del_nbytes (nbytes))
                break;
        }

         if (CLine == save_line) break;
         (void) jed_down(1);
    }

    line->next = NULL;

    Rectangle_Buffer->beg  = beg;
    Rectangle_Buffer->line = beg;
    Rectangle_Buffer->end  = line;
    Rectangle_Buffer->point = 0;

    pop_spot();

    //  when cutting, position cursor on top left corner
    if (kill)
    {
        goto_line(&N_ROW);
        goto_column(&W_COL);
        return 0;
    }

    //restore_rectangle();

    return 0;
}

////
int copy_rectangle (void)
{
    return copy_or_cut_rectangle (0);
    // leave rectangle alone
    // just copy it to rectangle_buffer
}

////
int cut_rectangle (void)
{
    return copy_or_cut_rectangle (1);
    // copy a rectangle to buffer and delete it
}

////
int kill_rectangle (void)
{
   return copy_or_kill_rectangle (1);
}

////
void save_selection (void)
{
    int n, sstatus, tmp=0;
    int r, nl, nc, sl, sc;
    char *ddir, *ffile, *ffile1, *ffile2;

    if (CBuf->vis_marks != 1) return;
    //   nothing to do

    ffile = read_from_minibuffer
    ("Save all LINES of block to file: ", " ", CBuf->dir, &n);
    //  entire lines will be saved
    //  even if only a portion of them is in the selection

    if ((ffile == NULL) || (*ffile == 0)) return;
    //  nothing entered

    if (-1 == jed_dirfile_to_dir_file (ffile, &ddir, &ffile1))
    {
        //  separate dir and file, no validation
        SLfree (ffile);
        flush_message("Unable to extract dir and filename");
        return;
    }

    if (bufferp(ffile1))
    {
        //  name entered is already in use in a buffer
        flush_message
        ("File is already associated to a buffer, choose different name.");
        return;
    }

    push_spot();

    RestoreHilight = CBuf->blocktype;
    r = RestoreHilight;

    //  rectangle
    if (CBuf->blocktype==1)
        check_rectangle();
    else
        check_stream();

    sstatus = file_status (ffile);

    if (sstatus == 2)
        jed_verror ("This is a directory!");

    else if (sstatus == -1)
        jed_verror ("File exists and is not accessible");

    else if (sstatus == -2)
        jed_verror ("Invalid path");

    else if (sstatus == -3)
        jed_verror ("Undetermined error");

    else if (sstatus == 0)
    {
        if (file_status (ddir) == 2) tmp = write_region (ffile);
        //  write attempt only if dir is valid
    }

    else if (sstatus == 1)
    {
        int ch;

        ch = jed_get_mini_response
        ("File exists: append/overwrite/cancel? (a/o/c) ");

        if (ch=='a' || ch=='A') append_to_file (ffile);

        else if (ch=='o' || ch=='O') tmp=write_region (ffile);

        else
        {
            if (r==2) stream_to_lineblock();
            pop_spot();
            return;
        }
    }

    if (sstatus == 0 && tmp < 1)
    {
        ffile2 = strcat (ffile," - Check path, saving failed!");
        flush_message(ffile2);
    }
    else if (tmp)
        flush_message("Done");

    pop_spot ();

    RestoreHilight = r;

    if (r==1)
        restore_rectangle();
    else
        restore_selection();
}
////

/*{{{ User Mark Functions */

typedef struct
{
    Mark m;                             /* MUST be the first */
    Buffer *b;
}
User_Mark_Type;

static void free_user_mark (SLtype type, VOID_STAR um_alias)
{
   Mark *m, *m1;
   Buffer *b;
   User_Mark_Type *um;

   (void) type;
   um = (User_Mark_Type *) um_alias;
   m1 = &um->m;

   /* The mark is only valid if the buffer that it was created for still
    * exists.
    */
   if ((m1->flags & MARK_INVALID) == 0)
     {
        /* Unlink the mark from the chain. */
        b = um->b;
        m = b->user_marks;
#if JED_HAS_LINE_MARKS
        if (m1->flags & JED_LINE_MARK)
          touch_screen ();
#endif
        if (m == m1)    b->user_marks = m1->next;
        else
          {
             while (m->next != m1) m = m->next;
             m->next = m1->next;
          }
     }

   SLfree ((char *)um);
}

void free_user_marks (Buffer * b)
{
   Mark * m = b->user_marks;

   while (m != NULL)
     {
        m->flags |= MARK_INVALID;
        m = m->next;
     }
}

static int mark_valid (Mark * m)
{
   if (m->flags & MARK_INVALID)
     {
        ////msg_error ("Mark is invalid.");
        return 0;
     }
   return 1;
}

static SLang_MMT_Type *pop_valid_user_mark (User_Mark_Type **ump)
{
   SLang_MMT_Type *mmt;
   User_Mark_Type *um;

   *ump = NULL;

   if (NULL == (mmt = SLang_pop_mmt (JED_MARK_TYPE)))
     return NULL;

   um = (User_Mark_Type *) SLang_object_from_mmt (mmt);

   if (0 == mark_valid (&um->m))
     {
        SLang_free_mmt (mmt);
        return NULL;
     }

   *ump = um;
   return mmt;
}

int jed_move_user_object_mark (SLang_MMT_Type *mmt)
{
   User_Mark_Type *um;
   Mark * m;

   um = (User_Mark_Type *) SLang_object_from_mmt (mmt);
   m = &um->m;

   if (!mark_valid (m)) return 0;

   if (CBuf != um->b)
     {
        ////msg_error ("Mark not in buffer.");
        message ("Mark not in buffer.");
        return 0;
     }

   m->line = CLine;
   m->point = Point;
   m->n = LineNum + CBuf->nup;
   return 1;
}

void move_user_mark (void)
{
   SLang_MMT_Type *mmt;

   if (NULL == (mmt = SLang_pop_mmt (JED_MARK_TYPE)))
     return;

   (void) jed_move_user_object_mark (mmt);
   SLang_free_mmt (mmt);
}


static int x_user_mark_fun (int (*xfun)(Mark *))
{
    SLang_MMT_Type *mmt;
    User_Mark_Type *um;
    int ret = -1;

    if (NULL == (mmt = pop_valid_user_mark (&um)))
        return -1;

    if (CBuf != um->b) msg_error ("Mark not in buffer.");
    else ret = (*xfun) (&um->m);

    SLang_free_mmt (mmt);
    return ret;
}

/* It is up to calling routine to ensure that mark is in buffer. */
////static int is_mark_in_narrow (Mark *m)
////int jed_is_user_mark_in_narrow (void)

void goto_user_mark (void)
{
   (void) x_user_mark_fun (jed_goto_mark);
}

SLang_MMT_Type * jed_make_user_object_mark (void)
{
   User_Mark_Type *um;
   SLang_MMT_Type *mmt;
   Mark *m;

   if (NULL == (um = (User_Mark_Type *) jed_malloc0 (sizeof(User_Mark_Type))))
     return NULL;

   if (NULL == (mmt = SLang_create_mmt (JED_MARK_TYPE, (VOID_STAR) um)))
     {
        SLfree ((char *) um);
        return NULL;
     }

   m = &um->m;

   jed_init_mark (m, 0);

   m->next = CBuf->user_marks;

   CBuf->user_marks = m;

   um->b = CBuf;

   return mmt;
}

void create_user_mark (void)
{
   SLang_MMT_Type * mmt;

   if (NULL != (mmt = jed_make_user_object_mark ()))
     {
        if (-1 == SLang_push_mmt (mmt))
          SLang_free_mmt (mmt);
     }
}

char * user_mark_buffer (void)
{
   SLang_MMT_Type *mmt;
   User_Mark_Type *um;
   char *s;

   if (NULL == (mmt = pop_valid_user_mark (&um)))
     return "";

   s = um->b->name;

   SLang_free_mmt (mmt);
   return s;
}

static int
user_mark_bin_op_result (int op, SLtype a, SLtype b,
                         SLtype *c)
{
   (void) a; (void) b;
   switch (op)
     {
      default:
        return 0;

      case SLANG_GT:
      case SLANG_GE:
      case SLANG_LT:
      case SLANG_LE:
      case SLANG_EQ:
      case SLANG_NE:
        *c = SLANG_INT_TYPE;
        break;
     }
   return 1;
}

static int
user_mark_bin_op (int op,
                  SLtype a_type, VOID_STAR ap, unsigned int na,
                  SLtype b_type, VOID_STAR bp, unsigned int nb,
                  VOID_STAR cp)
{
   int *ic;
   unsigned int n, n_max;
   unsigned int da, db;
   SLang_MMT_Type **a, **b;

   (void) a_type;
   (void) b_type;

   if (na == 1) da = 0; else da = 1;
   if (nb == 1) db = 0; else db = 1;

   if (na > nb) n_max = na; else n_max = nb;

   a = (SLang_MMT_Type **) ap;
   b = (SLang_MMT_Type **) bp;

   ic = (int *) cp;

   for (n = 0; n < n_max; n++)
     {
        User_Mark_Type *ua, *ub;
        Buffer *ba, *bb;
        int pa, pb;
        unsigned int la, lb;

        ub = NULL;
        ba = bb = NULL;
        pa = pb = 0;
        la = lb = 0;

        if ((*a != NULL)
            && (NULL != (ua = (User_Mark_Type *) SLang_object_from_mmt (*a))))
          {
             la = ua->m.n;
             pa = ua->m.point;
             ba = ua->b;
          }

        if ((*b != NULL)
            && (NULL != (ub = (User_Mark_Type *) SLang_object_from_mmt (*b))))
          {
             lb = ub->m.n;
             pb = ub->m.point;
             bb = ub->b;
          }

        if ((ba == NULL) || (bb == NULL))
          {
             ic[n] = 0;
             a += da;
             b += db;
             continue;
          }

        switch (op)
          {
           case SLANG_GT:
             ic[n] = ((la > lb)
                      || ((la == lb) && (pa > pb)));
             break;

           case SLANG_GE:
             ic[n] = ((la > lb)
                      || ((la == lb) && (pa >= pb)));
             break;

           case SLANG_LT:
             ic[n] = ((la < lb)
                      || ((la == lb) && (pa < pb)));
             break;

           case SLANG_LE:
             ic[n] = ((la < lb)
                      || ((la == lb) && (pa <= pb)));
             break;

           case SLANG_EQ:
             ic[n] = ((ba == bb) && (la == lb) && (pa == pb));
             break;

           case SLANG_NE:
             ic[n] = ((ba != bb) || (la != lb) || (pa != pb));
             break;

           default:
             return 0;
          }

        a += da;
        b += db;
     }

   return 1;
}

#if JED_HAS_LINE_MARKS
void jed_create_line_mark (int *color)
{
   SLang_MMT_Type *mmt;
   User_Mark_Type *um;

   if (NULL == (mmt = jed_make_user_object_mark ()))
     return;

   um = (User_Mark_Type *) SLang_object_from_mmt (mmt);
   um->m.flags |= JED_LINE_MARK | (*color & MARK_COLOR_MASK);
   if (-1 == SLang_push_mmt (mmt))
     SLang_free_mmt (mmt);
}

#endif

#ifndef SLFUTURE_CONST
# define SLFUTURE_CONST
#endif

static int user_mark_sget (SLtype type, SLFUTURE_CONST char *name)
{
   SLang_MMT_Type *mmt;
   User_Mark_Type *um;
   int status;
   Buffer *buf;

   (void) type;

   if (NULL == (mmt = pop_valid_user_mark (&um)))
     return -1;

   buf = um->b;

   status = -1;
   if (0 == strcmp (name, "buffer_name"))
     status = SLang_push_string (buf->name);
   else
     SLang_verror (SL_NOT_IMPLEMENTED,
                   "Mark_Type.%s is invalid", name);

   SLang_free_mmt (mmt);
   return status;
}

int register_jed_classes (void)
{
   SLang_Class_Type *cl;

   cl = SLclass_allocate_class ("Mark_Type");
   if (cl == NULL) return -1;
   (void) SLclass_set_destroy_function (cl, free_user_mark);

   (void) SLclass_set_sget_function (cl, user_mark_sget);

   if (-1 == SLclass_register_class (cl, JED_MARK_TYPE, sizeof (User_Mark_Type), SLANG_CLASS_TYPE_MMT))
     return -1;

   if (-1 == SLclass_add_binary_op (JED_MARK_TYPE, JED_MARK_TYPE, user_mark_bin_op, user_mark_bin_op_result))
     return -1;

   return 0;
}

////void jed_widen_whole_buffer (Buffer *b)
////#if JED_HAS_SAVE_NARROW
////int jed_count_narrows (void)

unsigned int jed_count_lines_in_region (void)
{
   Mark *m;
   unsigned int n0, n1;

   m = CBuf->marks;
   if (m == NULL)
     return 0;

   n0 = m->n - CBuf->nup;
   n1 = LineNum;

   if (n0 > n1)
     return 1 + (n0 - n1);

   return 1 + (n1 - n0);
}
